package com.simonstl.midp.xml;

/*
The contents of this file are subject to the Mozilla Public License  Version 1.1 (the "License"); you may not use this file except in  compliance with the License. You may obtain a copy of the License at 
http://www.mozilla.org/MPL/ 
Software distributed under the License is distributed on an "AS IS"  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  License for the specific language governing rights and limitations under  the License. 

The Original Code is available at  http://simonstl.com/projects/fragment/original. 

The Initial Developer of the Original Code is Simon St.Laurent. Portions created by Simon St.Laurent are Copyright (C) 2001. All Rights Reserved. 

Contributor(s): 
*/

import com.simonstl.midp.tam.*; 

import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Hashtable;
import java.util.Vector;

/**
 *
 * A simple implementation of an XML parser using the TAM interface.
 *
 * Note that this parser leaves all DOCTYPE processing to the application, 
 * and that it can apply but not recognize namespace declarations.
 *
 * @author  simonstl
 */
public class TAMParserImpl implements TamParser {
    
    TamHandler handler;
    private static final int UNKNOWN=0;
    
    private static final int ELEMENT_START=1;
    private static final int ELEMENT_END=2;
    private static final int ELEMENT_SPACE=3;
        
    private static final int ATT_NAME=4;
    private static final int ATT_VALUE=5;
    private static final int ATT_EQUALS=6;
    
    private static final int PI_TARGET=8;
    private static final int PI_SPACE=9;
    private static final int PI_DATA=10;
    private static final int PI_QUESTION=11;
    
    private static final int CDATA_START=16;
    private static final int CDATA_CONTENT=17;
    private static final int CDATA_END=18;

    private static final int DOCTYPE=32;
    private static final int DOCTYPE_END_SQUARE=33;
        
    private static final int LESSTHAN=64;
    private static final int LESSTHAN_BANG=65;
        
    private static final int COMMENT_TEXT=128;
    private static final int COMMENT_FIRST=129;
    private static final int COMMENT_DASH=130;
    private static final int COMMENT_DASH_DASH=131;
    
    private static final int AMP=256;
    private static final int AMP_SHARP=257;
    private static final int AMP_SHARP_X=258;
    private static final int HEX_REF=259;
    private static final int DEC_REF=260;
    private static final int ENTITY=264;

    boolean quoteType=true; //true = "", false = '
    
    int size=128;
    //if reading is true, then we haven't reached the end of the buffer
    boolean reading=true;
    
    //if halt is true we've hit a serious error
    boolean halt=false;
    
    //if skipChar is true, we skip the rest of processing for this character
    boolean skipChar=false;
    
    //if targetSpace is true, we've encountered the first space in a PI
    boolean targetSpace=false;
    
    //whether or not we've finished the element name
    boolean elementNameEstablished=false;
    
    //at the colon or not yet
    boolean colonInName=false;
    
    //the main buffer for reading
    char[] buf=new char[size];
    
    //the internal storage buffer for content
    StringBuffer current=new StringBuffer();
    
    //the internal storage buffer for entities
    StringBuffer entBuffer=new StringBuffer();
    
    //int for where we were before entitiy
    int beforeEntity=UNKNOWN;
    
    //if putInBuffer is false, the current character doesn't go in the buffer
    boolean putInBuffer=true;
    
    //if pastXMLDecl is true, we've encountered something beyond where the XML decl is allowed
    boolean pastXMLDecl=false;
    
    //temporary holder for prefix info
    String prefix="";
    
    //temporary holder for element info
    TAMElement element=new TAMElement();
    
    //temporary holder for attribute info
    TAMAttribute attribute=new TAMAttribute();
    
    //PI target and data
    String target="";
    String data="";
    
    //Entities from application
    Hashtable entities=new Hashtable();
    
    //QNames for elements, atts
    Hashtable qNames=new Hashtable();
    
    //Local names for elements, atts
    Hashtable names=new Hashtable();
    
    //prefixes for elements, atts
    Hashtable prefixes=new Hashtable();
    
    //namespaces
    TAMNSContext namespaces=new TAMNSContext();
    
    //prior elements
    Vector elementStack=new Vector();
    
    //line-end normalization
    boolean xD=false;
    
    //if in the DOCTYPE, are we in the internal subset?
    boolean internalSubset=false;
    
    /** Creates a new instance of TAMParserImpl */
    public TAMParserImpl() {
        reset();
    }
    
    public void setMaxBuf(int _size){
       if (size>64) {size=_size;}
    }
    
    public TamHandler getHandler() {
        return handler;
    }
    
    public void setHandler(TamHandler _handler) {
        handler=_handler;
    }
    
    public void parse(InputStream in) throws TAMException {
        
        InputStreamReader r=new InputStreamReader(in);
        //if overflow is true, backup buffer has contents
        boolean putInBuffer=true;
        
        
        
        int state=UNKNOWN;
        int bigPosition=0;
        int numChars=0;

        
        while (reading) {
            //fill buffer
            try {
               numChars=r.read(buf, bigPosition, size);

               if (numChars<size) {reading=false;}
            
              //read through bytes
              //key characters are <, >, &, 
            
               for (int i=0;i<numChars;i++){
                   char currentChar=buf[i];
                   putInBuffer=true;
//reportState("Start", state, i, currentChar, putInBuffer);      
                    if (currentChar=='&'){
                       if ((state==UNKNOWN) || (state==ATT_VALUE)) {
                           beforeEntity=state;
                           state=AMP;
                           putInBuffer=false;
                           skipChar=true;
                       } else if ((state==COMMENT_TEXT) || (state==PI_DATA) || ((state==DOCTYPE) && internalSubset)) {
                           //everything's okay, so let it go through
                       } else {
                           //yecch... must be someplace illegal.
                           fatalError("Inappropriate ampersand");
                       }
                    }
                   if (currentChar=='<') {
                       if (!((state==UNKNOWN) || (state==COMMENT_TEXT) || (state==PI_DATA) || ((state==DOCTYPE) && internalSubset))) {

                           fatalError("Too many <s encountered");
                       } else {
                           if (state==UNKNOWN) {
                               if (current.length()>0) {
                                   reportCharacters();
                                   pastXMLDecl=true;
                               }
                               state=LESSTHAN;
                               putInBuffer=false;
                           }
                       } 
                   } else {   
//begin handlers for particular states-----------------------------------------------------
                       if ((state==UNKNOWN) || (state==ATT_VALUE)) {
                           //line-end normalization
                         if ((currentChar==13) || (currentChar==10)){
                           if ((currentChar==13) && (!xD)){
                               xD=true;
                               putInBuffer=false;
                           } else if (currentChar==13){
                               current.append((char)10);
                               putInBuffer=false;
                           } else if (xD) {
                               current.append((char)10);
                           } else if (currentChar==10) {
                               putInBuffer=true;
                               xD=false;
                           } 
                         }
                       }
                       if (state==ENTITY) {
                           if (currentChar==';'){
                               //need to create new char from 
                               //big problem - J2ME doesn't support hex->int
                               

                               String entName=entBuffer.toString();
                               String entValue=(String) entities.get(entName);
                               if (entValue==null){
                                   reportSkippedEntity(entName);
                               } else {
                                   current.append(entValue);
                               }
                               
                               
                               state=beforeEntity;
                               entBuffer=new StringBuffer();
                           } else {
                              entBuffer.append(currentChar);
                           }
                           putInBuffer=false;
                       }
                              
                       if ((state==HEX_REF) || (state==DEC_REF)){

                           if (currentChar==';'){
                               //need to create new char from
                               
                               
                               int value=0;

                               if (state==HEX_REF){
                                   value=hexRefToDec();
                               } else {
                                  value=Integer.parseInt(entBuffer.toString());   
                               }
                               state=beforeEntity;
                               //watch for surrogates!
                               
                               if (value<65535){
                                   //regular char, one char output
                                   currentChar=(char)value;
                                   
                               } else {
                                   //surrogate pair handling
                                   int s=value-65546;
                                   int high=(s / 1024) + 55296;
                                   int low=(s % 1024) + 56320;
                                   current.append((char)high);
                                   currentChar=(char)low;                                  
                               }
                               
                               putInBuffer=true;
                               entBuffer=new StringBuffer();
                           } else {
                              entBuffer.append(currentChar);
                              putInBuffer=false;

                           }
                           
                       }
                       
                       if (state==AMP_SHARP){
                           
                               if (currentChar=='x'){
                                   state=HEX_REF;

                               } else {
                                   state=DEC_REF;
                                   entBuffer.append(currentChar);
                               }
                           
                           putInBuffer=false;
                       }
                       
                       if (state==AMP){

                         if(!skipChar){
                           if (currentChar=='#'){
                             state=AMP_SHARP;   
                           } else {
                            //add check for acceptable chars - someday
                             state=ENTITY;
                             entBuffer.append(currentChar);
                           }
                           
                           
                           putInBuffer=false;
                         } else {skipChar=false;}
                       }
                       
                       if (state==DOCTYPE_END_SQUARE){
                           if (currentChar=='>'){
                               reportDOCTYPE();
                               state=UNKNOWN;
                               putInBuffer=false;
                           }
                       }
                       
                   if (state==DOCTYPE) {
                       //we're only looking for >, [, or ]
                       if (!internalSubset){
                         if (currentChar=='>'){
                             reportDOCTYPE();
                             state=UNKNOWN;
                             putInBuffer=false;
                         } else if (currentChar=='[') {
                             internalSubset=true;
                         }
                       } else {
                           if (currentChar==']'){
                               state=DOCTYPE_END_SQUARE;
                           }
                       }
                   }
                   
                   if (state==COMMENT_DASH_DASH) {
                      if (currentChar=='>'){
                           reportComment();
                           state=UNKNOWN;
                           putInBuffer=false;
                      } else {
                          warning("Misformed comment --");
                          current.append("--");
                          state=COMMENT_TEXT;
                      }    
                   }
                   
                   if (state==COMMENT_DASH) {
                      if (currentChar=='-'){
                          state=COMMENT_DASH_DASH;
                          putInBuffer=false;
                      } else {
                          current.append('-');
                          state=COMMENT_TEXT;
                      }
                   }
                       
                   if (state==COMMENT_TEXT) {
                      if (currentChar=='-'){
                          state=COMMENT_DASH;
                          putInBuffer=false;
                      }
                   }
                       
                   if (state==COMMENT_FIRST) {
                       if (currentChar=='-'){
                          state=COMMENT_TEXT;
                          putInBuffer=false;
                       } else {
                         state=UNKNOWN;
                         fatalError("Botched comment");
                       }
                   }

                   if (state==LESSTHAN_BANG){
                       if (currentChar=='-'){
                          state=COMMENT_FIRST;
                          putInBuffer=false;
                       } else {
                           state=DOCTYPE;
                       }
                   }
        

                   if (state==PI_QUESTION){
                       if (currentChar=='>'){
                           data=current.toString();
                           reportPI();
                           state=UNKNOWN;
                           putInBuffer=false;
                           current=new StringBuffer();
                      } else if (targetSpace) {
                           current.append('?');
                           state=PI_DATA;
                       } else {
                           fatalError("Faulty PI with target: "+target);
                           
                       }
                   }
                   
                   if (state==PI_DATA) {
                      if (currentChar=='?'){
                          putInBuffer=false;
                          state=PI_QUESTION;
                      }
                   }
                   
                   if (state==PI_SPACE) {
                       if (isSpace(currentChar)){
                          putInBuffer=false;   
                       } else {
                           state=PI_DATA;
                       }
                   }
                       
                   if (state==PI_TARGET){
                       if (isSpace(currentChar)){
                           state=PI_SPACE;
                           putInBuffer=false;
                           target=current.toString();
                           current=new StringBuffer();
                           targetSpace=true;
                       } else if (currentChar=='?'){
                           state=PI_QUESTION;
                           putInBuffer=false;
                           target=current.toString();
                           current=new StringBuffer();
                       }
                   }
                   
                   if (state==ATT_VALUE){
                     if (( (currentChar=='"') && (quoteType) ) || 
                     ( (currentChar=='\'') && (!(quoteType)) )) {
                         state=ELEMENT_SPACE;
                         putInBuffer=false;
                         String value=current.toString();
                         if (attribute.getPrefix().equals("xmlns")){
                           namespaces.declarePrefix(attribute.getLocalName(), value);
                         } else if ( (attribute.getPrefix().equals("")) && (attribute.getLocalName().equals("xmlns")) ){
                           namespaces.declarePrefix("", value);
                         }
                         attribute.setValue(value);
                         attribute.setQuotes(quoteType);
                         element.addAttribute(attribute);
                         attribute=new TAMAttribute();
                         current=new StringBuffer();
                         skipChar=true;
                     }  
                         
                     
                   }

                   if (state==ATT_EQUALS){
                     if (isSpace(currentChar)){
                         putInBuffer=false;
                     } else if (currentChar=='"') {
                         putInBuffer=false;
                         boolean quoteType=true;
                         state=ATT_VALUE;
                     } else if (currentChar=='\'') {
                         putInBuffer=false;
                         boolean quoteType=false;
                         state=ATT_VALUE;
                     }
                     
                   }

                   
                   if (state==ATT_NAME){
                     if (currentChar=='='){
                         state=ATT_EQUALS;
                         putInBuffer=false;
                         String name=normName(current.toString());
                         attribute.setPrefix(prefix);
                         //URI will be set when element reported
                         attribute.setLocalName(name);
                         current=new StringBuffer();
                         colonInName=false;
                         prefix="";
                     } else if (currentChar==':') {
                           if (colonInName){
                               fatalError("Too many colons in name");
                           } else {
                               prefix=normPrefix(current.toString());
                               colonInName=true;
                               current=new StringBuffer();
                               putInBuffer=false;
                           }
                     } else if (isSpace(currentChar)){
                         putInBuffer=false;
                     }
                   }
                  
                   if (state==ELEMENT_END){
                     //check for space or >, otherwise report end of element 
                     if (currentChar=='>') {
                          //report endElement - strip space from end of element if nec
                         putInBuffer=false;
                         String name=normName(current.toString());
                         element.setPrefix(prefix);
                         //URI will be set when element reported
                         element.setLocalName(name);
                         reportElementEnd();
                         state=UNKNOWN;
                         colonInName=false;
                         prefix="";
                       } else if (currentChar==':') {
                           if (colonInName){
                               fatalError("Too many colons in name");
                           } else {
                               prefix=normPrefix(current.toString());
                               colonInName=true;
                               current=new StringBuffer();
                               putInBuffer=false;
                           }
                       }
                   }
                  
                   if (state==ELEMENT_SPACE){
                    if (!skipChar) {
                     //check for space or >, otherwise start attribute 
                       if (isSpace(currentChar)) {
                         putInBuffer=false;
                       } else if (currentChar=='>') {
                          putInBuffer=false;
                          if (!elementNameEstablished) {
                           String name=normName(current.toString());
                           element.setPrefix(prefix);
                           //URI will be set when element reported
                           element.setLocalName(name);
                          }
                           reportElementStart();
                           state=UNKNOWN;
                       } else if (currentChar=='/') {
                          //empty element
                           putInBuffer=false;
                           element.setIsEmpty(true);
                       } else {
                           current=new StringBuffer();
                           state=ATT_NAME;
                       }
                    } else {
                        skipChar=false;
                    }
                   }
                   
                   if (state==ELEMENT_START){
                     //check for whitespace, save  
                       if (isSpace(currentChar)){
                           state=ELEMENT_SPACE;
                           String name=normName(current.toString());
                           element.setPrefix(prefix);
                           //URI will be set when element reported
                           element.setLocalName(name);
                           elementNameEstablished=true;
                           colonInName=false;
                           prefix="";
                        //put element name somewhere
                           putInBuffer=false;
                       } else if (currentChar==':') {
                           if (colonInName){
                               fatalError("Too many colons in name");
                           } else {
                               prefix=normPrefix(current.toString());
                               colonInName=true;
                               current=new StringBuffer();
                               putInBuffer=false;
                           }
                           
                       } else if (currentChar=='/') {
                          //empty element
                           putInBuffer=false;
                           element.setIsEmpty(true);
                       } else if (currentChar=='>') {
                          //element name=buffer
                           putInBuffer=false;
                   
                           String name=normName(current.toString());
                           element.setPrefix(prefix);
                           //URI will be set when element reported
                           element.setLocalName(name);
                           reportElementStart();
                           elementNameEstablished=true;
                           colonInName=false;
                           prefix="";
                           state=UNKNOWN;
                       }
                   } 
                   
                   if (state==LESSTHAN){
                      if (currentChar=='?'){
                          state=PI_TARGET;
                          targetSpace=false;
                          putInBuffer=false;
                      } else if (currentChar=='!') {
                          state=LESSTHAN_BANG;
                          putInBuffer=false;
                          pastXMLDecl=true;
                      } else if (currentChar=='/') {
                          elementNameEstablished=false;
                          state=ELEMENT_END;
                          putInBuffer=false;
                          pastXMLDecl=true;
                      } else if (isSpace(currentChar)) {
                          //whitespace - add LF, CR
                          fatalError("Whitespace at wrong place");
                      } else {
                          state=ELEMENT_START;
                          elementNameEstablished=false;
                          //setup for element
                          putInBuffer=true;
                          pastXMLDecl=true;
                      }
                   }
                    
                   
            
                } //end if char=<
                if (putInBuffer){
                 current.append(currentChar);   
                }
//reportState("End  ", state, i, currentChar, putInBuffer);  
                if (halt) {break;}
              } //end for
               
              //handle end-of-buffer situation
               
            } catch (Exception e) {
               System.out.println("blam!" + e);
               reading=false;
            }
            

            
        }//end while (reading)
    }
    
    protected boolean isSpace(char c){
        boolean returnVal=false;
        
        if ((c==' ') || (c==0x9) || (c==0xA) || (c==0xD)) {returnVal=true;}
        
        return returnVal;
    }
    
    protected int hexRefToDec(){
       //required because J2ME doesn't convert 0xABCD to integers
       int value=0;
       int place=1;
       int digit=0;
       for(int i=entBuffer.length(); i>0; i--){
           char digitChar=entBuffer.charAt(i-1);
           
           if ( (digitChar>='0') && (digitChar<='9') ){
               digit=(int)digitChar-48;
           } else if ( (digitChar>='A') && (digitChar<='F') ){
               digit=(int)digitChar-55;
               
           }
           
           value=value + (digit * place);
           
           place=place * 16;
       }
       return value;   
    }
    

    
        
    
    //fatal errors are reported and processing halts
    protected void fatalError(String s) throws TAMException{
        TAMException e=new TAMException(s);
        try {
          handler.fatalError(e);
        } catch (TAMException ex) {
          //oy, we're hosed
            throw e;
        }
        halt=true;
        reading=false;
    }
    
    //errors are reported and processing halts
    protected void error(String s) throws TAMException{
        TAMException e=new TAMException(s);
        try {
           handler.error(e);
        } catch (TAMException ex) {
          //oy, we're hosed
            throw e;
        }
        reading=false;
    }
    
    //warnings are reported and processing continues
    protected void warning(String s)  throws TAMException{
        TAMException e=new TAMException(s);
        try {
            handler.warning(e);
        } catch (TAMException ex) {
          //oy, we're hosed
            throw e;
        }
    }
    
    protected void reportElementStart() throws TAMException {
        element.setUri(namespaces.getUri(element.getPrefix()));
 //need to pass this a namespace object
        element.setAttUris(namespaces);
        String prefix=element.getPrefix();
        String name=element.getLocalName();
        
        handler.startElement(element.getUri(), name, prefix, element.getAttributes() );
        namespaces.pushLevel();
        elementStack.addElement(prefix+":"+name);
        
        if (element.getIsEmpty()){
            
             handler.endElement(element.getUri(), name, prefix);
             namespaces.popLevel();
             String qName=prefix+":"+name;
        
        String startName=(String) elementStack.lastElement();
        
        if (! (qName.equals(startName))){ fatalError(qName+" does not match "+startName); }
        
        int size=elementStack.size() -1;
//System.out.println(size);
        elementStack.removeElementAt(size);
        }
        
        
        element=new TAMElement();
        current=new StringBuffer();
        
    }
    
    protected void reportCharacters() throws TAMException {
        int length=current.length();
        char[] ch=new char[length];
        current.getChars(0, length, ch, 0);
        
        handler.characters(ch, 0, length);
        
        current=new StringBuffer();
        
    }
    protected void reportElementEnd() throws TAMException {
        String prefix=element.getPrefix();
        String name=element.getLocalName();
        
        String uri=(namespaces.getUri(prefix));
        namespaces.popLevel();

        
        String startName=(String) elementStack.lastElement();
        String qName=prefix+":"+name;
        if (! (qName.equals(startName))){
            
            fatalError(qName+" does not match "+startName); 
        }
        
        int size=elementStack.size()-1;
//System.out.println(size);
        elementStack.removeElementAt(size);
        
        handler.endElement(uri, name, prefix );
        
        element=new TAMElement();
        current=new StringBuffer();
        
    }
    
    protected void reportPI() throws TAMException {
        if (!pastXMLDecl && target.equals("xml")){

            //parse XMLDecl
            String version="1.0";
            String encoding="UTF-8";
            String standalone="yes";
            //need to implement parse!
            
            handler.XMLDecl(version, encoding, standalone);
        } else {
           handler.processingInstruction(target, data);
           target=new String();
           data=new String();
        }
        pastXMLDecl=true;
    }
    
    protected void reportSkippedEntity(String entName) throws TAMException {
       
           handler.skippedEntity(entName);
           
        pastXMLDecl=true;
    }
    
    protected void reportComment() throws TAMException {
        int length=current.length();
        char[] ch=new char[length];
        current.getChars(0, length, ch, 0);
        
        handler.comment(ch, 0, length);
        
        current=new StringBuffer();
    }
    
    protected void reportDOCTYPE() throws TAMException{
        int length=current.length();
        char[] ch=new char[length];
        current.getChars(0, length, ch, 0);
        
        handler.doctype(ch, 0, length);
        
        current=new StringBuffer();
    }
    
    protected void reportState(String label, int state, int num, char currentChar, boolean buffer){
        System.out.println(label+" - "+" Pos: "+num+" State:"+state+" Char: "+currentChar+" Inc: "+buffer +" Buf: "+current +" EntBuf:"+entBuffer);
    }
    
    
    /**
     */
    public void setFeature(String name, boolean value) {
    }
    
    public boolean getFeature(String name) {
        return false;
    }
    
    public void reset(){
        entities.clear();
        entities.put("lt", "<");
        entities.put("gt", ">");
        entities.put("amp", "&");
        entities.put("apos", "'");
        entities.put("quot", "\"");
    }
    
    public void setEntity(String name, String value) throws TAMException{
        entities.put(name, value);
    }
   
    protected String normQName(String qName){
        String returnVal=qName;
        String qValue=(String) qNames.get(qName);
         if (qValue==null){
             qNames.put(qName, qName);
         } else {
             returnVal=qValue;
         }
        return returnVal;
    }
 
    protected String normName(String name){
        String returnVal=name;
        String value=(String) names.get(name);
         if (value==null){
             names.put(name, name);
         } else {
             returnVal=value;
         }
        return returnVal;
    }

    protected String normPrefix(String name){
        String returnVal=name;
        String value=(String) prefixes.get(name);
         if (value==null){
             names.put(name, name);
         } else {
             returnVal=value;
         }
        return returnVal;
    }
    
    
   public void setNamespace(String prefix, String nsURI) throws TAMException{
       namespaces.declarePrefix(prefix,nsURI);
   }
   
}
